#define _GNU_SOURCE /* needed to get the defines */
#include <fcntl.h>  /* in glibc 2.2 this has the needed values defined */
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>


static char *progname = NULL;
static char *dir_name = ".";
static char *log_name = NULL;
static int   trunc_log = 0;

static int parse_options(int argc, char **argv);


static void handler(int sig, siginfo_t *si, void *data)
{
    FILE *stream;
    time_t t;

    stream = fopen(log_name, "a");
    if ( !stream )
        exit(1);
    time(&t);
    fprintf( stream, "Directory `%s' accessed on   %s",
             dir_name, asctime(localtime(&t)) );
    fclose(stream);

    return;
}


int main(int argc, char **argv)
{
    struct sigaction act;
    pid_t pid;
    int fd, i;

    /* Process command line arguments */
    progname = *argv;
    i = parse_options(argc, argv);
    if ( i == 0 )
        return 1;
    else if ( i == 1 )
        return 0;

    pid = fork();
    switch ( pid ) {
    case 0: /* inside new process */
        /* Break off into separate process group */
        setsid();
        /* Set handler for a real-time signal */
        act.sa_sigaction = handler;
        sigemptyset(&act.sa_mask);
        act.sa_flags = SA_SIGINFO;
        sigaction(SIGRTMIN, &act, NULL);
        /* We want to receive SIGRTMIN signal for access notification */
        fd = open(dir_name, O_RDONLY);
        if ( fd == -1 ) {
            fprintf(stderr,
                    "%s: error opening directory `%s' for reading\n",
                    progname, dir_name);
            exit(1);
        }
        if ( fcntl(fd, F_SETSIG, SIGRTMIN) == -1 ||
             fcntl(fd, F_NOTIFY, DN_ACCESS | DN_MULTISHOT)== -1 ) {
            fprintf(stderr,
                    "%s: error setting up watch for directory `%s'\n",
                    progname, dir_name);
            exit(1);
        }
        /* Close all standard streams--we won't need them */
        fclose(stdin);
        fclose(stdout);
        fclose(stderr);

        if ( trunc_log )
            truncate(log_name, 0);

        /* We will now be notified if any file was accessed */
        while ( 1 )
            pause();

        break;
    case 1: /* no fork for us */
        fprintf(stderr, "%s: error--unable to fork\n", progname);
        exit(1);
        break;
    default: /* inside parent--report new process's ID */
        printf("%d\n", pid);
        break;
    }

    return 0;
}



static char *errUnknownOption = "%s: error--unknown option `-%c'\n";
static char *errUnknownOptionChar = "%s: error--unknown option character `\\x%x'\n";
static char *errArgumentReq   = "%s: error--option `-%c' requires an argument (%s)\n";
static char *errLogName = "%s: error--log file name not specified\n";

static char *usehelp = "Try `%s -h' for help\n";
static char *usage   = "\n\
Usage: %s [-h] [-d DIR_NAME] [-t] LOG_NAME\n\
\n\
Watches files in the directory for access, and logs access timestamp to the\n\
log file\n\
\n\
-h            display this help message\n\
-d DIR_NAME   directory name to watch (default: current directory)\n\
-t            truncate log file on startup\n\
\n\
LOG_NAME is name of the file to save the log in.  If it is in the directory you\n\
are watching, several access notifications will get logged.\n\
\n";

/* Return codes:
 * 0    - error, exit(1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
static int parse_options(int argc, char **argv)
{
    char c, *ptr;
    int i;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":d:th")) != -1 )
        switch ( c ) {
        case 'd':
            dir_name = optarg;
            break;
        case 't':
            trunc_log = 1;
            break;
        case 'h':
            printf(usage, progname);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            return 0;
            break;
        case ':': { /* missing option argument */
            char *msg = NULL;

            switch ( optopt ) {
            case 'd':
                msg = "directory name to watch";
                break;
            default:
                assert(0); /* should never get here */
                abort();
                break;
            }
            fprintf(stderr, errArgumentReq, progname, optopt, msg);
            fprintf(stderr, usehelp, progname);
            return 0;
            break;
        }
        default:
            assert(0); /* should never happen */
            abort();
            break;
        }

    if ( optind >= argc ) {
        fprintf(stderr, errLogName, progname);
        return 0;
    }

    log_name = *(argv + optind);

    return -1;
}



/*
 * Local variables:
 * compile-command: "gcc -o satan satan.c"
 * End:
 */
